from ..pipeline import AgentCallContext, AgentTranscriptionOptions, VoicePipelineAgent

AssistantTranscriptionOptions = AgentTranscriptionOptions
AssistantCallContext = AgentCallContext
VoiceAssistant = VoicePipelineAgent

__all__ = [
    "AssistantTranscriptionOptions",
    "AssistantCallContext",
    "VoiceAssistant",
]
